from collections import Counter
import io
import zipfile

from immunedb.common.models import Clone, CloneStats, Sample
from immunedb.exporting.tsv_writer import StreamingTSV, write_tsv
from immunedb.util.log import logger
from immunedb.util.lookups import aas_from_nts


def get_clone_features(session):
    return {c.id: (c.v_gene, c.j_gene, c.cdr3_nt)
            for c in session.query(Clone.id, Clone.v_gene, Clone.j_gene,
                                   Clone.cdr3_nt)}


def get_sample_vdjtools(session, sample, min_clone_size, clone_features):
    writer = StreamingTSV(['count', 'freq', 'cdr3nt', 'cdr3aa', 'v', 'd', 'j'])

    sample_clones = Counter()
    stats = session.query(
        CloneStats.clone_id, CloneStats.total_cnt
    ).filter(
        CloneStats.sample_id == sample.id
    )

    for stat in stats:
        sample_clones[clone_features[stat.clone_id]] += stat.total_cnt

    total = sum(sample_clones.values())
    yield writer.writeheader()
    for key in sorted(sample_clones, key=sample_clones.get, reverse=True):
        counts = sample_clones[key]
        if counts < min_clone_size:
            continue
        v, j, cdr3_nt = key
        yield writer.writerow({
            'count': counts,
            'freq': counts / total,
            'cdr3nt': cdr3_nt,
            'cdr3aa': aas_from_nts(cdr3_nt),
            'v': v,
            'd': '.',
            'j': j,
        })


def write_vdjtools(session, args):
    clone_features = get_clone_features(session)
    for sample in session.query(Sample):
        logger.info('Exporting VDJTools format for sample {}'.format(
            sample.name))
        write_tsv(
            '{}.sample.txt'.format(sample.name),
            get_sample_vdjtools,
            session, sample, args.min_clone_size, clone_features
        )


def get_vjdtools_as_zip(session, min_clone_size=0):
    clone_features = get_clone_features(session)

    out = io.BytesIO()
    with zipfile.ZipFile(out, 'w') as fh:
        for sample in session.query(Sample):
            output = ''.join(get_sample_vdjtools(
                session, sample, min_clone_size, clone_features
            ))
            fh.writestr('{}.sample.txt'.format(sample.name), output)
    yield out.getvalue()



import csv
import dnautils
import sys

from immunedb.common.models import (Clone, NoResult, SampleMetadata, Sample,
                                    SampleStats, Sequence, SequenceCollapse)
from immunedb.identification.metadata import NA_VALUES
from immunedb.util.log import logger


def remove_duplicates(session, sample):
    logger.info('Removing duplicates from sample {}'.format(sample.id))
    all_seqs = session.query(Sequence).filter(
        Sequence.sample == sample
    ).order_by(
        Sequence.copy_number.desc()
    )

    buckets = {}
    for seq in all_seqs:
        key = (seq.v_gene, seq.j_gene, seq.cdr3_num_nts)
        buckets.setdefault(key, []).append(seq)

    for i, bucket in enumerate(buckets.values()):
        while len(bucket) > 0:
            larger = bucket.pop(0)
            for i in reversed(range(len(bucket))):
                smaller = bucket[i]
                if dnautils.equal(larger.sequence, smaller.sequence):
                    larger.copy_number += smaller.copy_number
                    session.delete(smaller)
                    del bucket[i]

    session.commit()


def update_metadata(session, args):
    SENTINEL = '__TEMP'  # Used to temporarily avoid duplicate name issues
    with open(args.new_metadata) as fh:
        reader = csv.DictReader(fh, delimiter='\t')
        new_meta = {l['name']: l for l in reader}

    # delete existing metadata
    sample_ids = {s.name: s.id for s in session.query(Sample).filter(
        Sample.name.in_(new_meta))}

    session.query(SampleMetadata).filter(
        SampleMetadata.sample_id.in_(sample_ids.values())
    ).delete(synchronize_session='fetch')

    ignore_fields = ['name', 'new_name', 'subject', 'file_name']
    for sample_name, row in new_meta.items():
        if sample_name not in sample_ids:
            logger.warning('No sample {} in database.  Ignoring.'.format(
                sample_name))
        sample_id = sample_ids[sample_name]
        logger.info('Updating metadata for {}'.format(row['name']))
        session.add_all([
            SampleMetadata(sample_id=sample_id, key=k, value=v)
            for k, v in row.items() if k not in ignore_fields and v not in
            NA_VALUES
        ])
        if row['new_name'] != row['name']:
            logger.info('  Updating sample name to {}'.format(row['new_name']))
            session.query(Sample).filter(Sample.name == row['name']).update({
                Sample.name: row['new_name'] + SENTINEL
            })

    logger.info('Verifying uniqueness')
    for sample in session.query(Sample).filter(
            Sample.name.like('%' + SENTINEL)):
        sample.name = sample.name[:-len(SENTINEL)]

    if session.query(Clone.id).filter(~Clone.tree.is_(None)).count() > 0:
        logger.warning('This database has at least one clonal lineage '
                       'constructed.  All lineages will need to be updated '
                       'to reflect the modified metadata.')
    session.commit()


def combine_samples(session, args):
    groups = {}

    for meta in session.query(SampleMetadata).filter(
            SampleMetadata.key == args.combine_field):
        groups.setdefault(meta.value, set()).add(meta.sample_id)
    all_subjects = set()
    for group_id, samples in groups.items():
        group_subs = session.query(Sample.subject_id).filter(
            Sample.id.in_(samples)
        ).group_by(Sample.subject_id)
        group_subs = [s.subject_id for s in group_subs]
        all_subjects.update(set(group_subs))
        if len(group_subs) > 1:
            logger.error('Cannot combine samples across subjects '
                         '(group "{}" has {} subjects)'.format(
                             group_id, len(group_subs)))
            sys.exit(1)

    all_samples = [s.id for s in session.query(Sample.id).filter(
        Sample.subject_id.in_(all_subjects))]

    logger.info('Resetting information for {} subjects'.format(
        len(all_subjects), len(all_samples)))
    logger.info('   Resetting collapsing')
    session.query(SequenceCollapse).filter(
        SequenceCollapse.sample_id.in_(all_samples)
    ).delete(synchronize_session=False)
    logger.info('   Resetting clones')
    session.query(Clone).filter(
        Clone.subject_id.in_(all_subjects)
    ).delete(synchronize_session=False)
    logger.info('   Resetting sample statistics')
    session.query(SampleStats).filter(
        SampleStats.sample_id.in_(all_samples)
    ).delete(synchronize_session=False)

    for group_id, samples in groups.items():
        final_sample_id = min(samples)
        logger.info('Combining {} samples into new sample "{}" (ID {})'.format(
            len(samples), group_id, final_sample_id))
        session.query(Sequence).filter(
            Sequence.sample_id.in_(samples)
        ).update({
            Sequence.sample_id: final_sample_id,
        }, synchronize_session=False)

        logger.info('Updating sample name and deleting empty samples')
        # collapse to one sample
        final_sample = session.query(Sample).get(final_sample_id)
        final_sample.name = group_id
        remove_duplicates(session, final_sample)

        logger.info('Moving noresults')
        session.query(NoResult).filter(
            NoResult.sample_id.in_(samples)
        ).update({
            'sample_id': final_sample_id
        }, synchronize_session=False)

        # delete the now-empty samples
        session.query(Sample).filter(
            Sample.id.in_(samples - set([final_sample_id]))
        ).delete(synchronize_session=False)

    session.commit()
    logger.info('Sequences successfully collapsed: please re-run '
                'immunedb_collapse and later pipeline steps.')



from collections import OrderedDict
from sqlalchemy.orm import joinedload

from immunedb.identification.genes import GeneName
from immunedb.common.models import (SampleMetadata, Sequence, SequenceCollapse,
                                    Subject)
from immunedb.util.log import logger
from immunedb.util.funcs import yield_limit
from immunedb.exporting.tsv_writer import StreamingTSV, write_tsv


mappings = {
    'airr': OrderedDict((
        ('sequence_id', 'seq_id'),
        ('sequence', 'original_sequence'),
        ('rev_comp', 'rev_comp'),
        ('productive', 'functional'),
        ('vj_in_frame', 'in_frame'),
        ('stop_codon', 'stop'),
        ('locus', lambda s: GeneName(s.v_gene).base),
        ('v_call', 'v_gene'),
        ('d_call', None),
        ('j_call', 'j_gene'),
        ('sequence_alignment', 'sequence'),
        ('germline_alignment', 'germline'),
        ('junction', 'cdr3_nt'),
        ('junction_aa', 'cdr3_aa'),
        ('v_score', 'v_match'),
        ('v_identity', lambda s: s.v_match / float(s.v_length)),
        ('v_cigar', 'v_cigar'),
        ('d_cigar', None),
        ('j_identity', lambda s: s.j_match / float(s.j_length)),
        ('j_score', 'j_match'),
        ('j_cigar', 'j_cigar'),
        ('v_sequence_start', 'seq_start'),
        ('duplicate_count', 'copy_number'),
        ('clone_id', 'clone_id'),

        # Extras
        ('clone_junction_nt', lambda s: s.clone.cdr3_nt if s.clone else ''),
        ('clone_junction_aa', lambda s: s.clone.cdr3_aa if s.clone else ''),
        ('sample', lambda s: s.sample.name),
        ('subject', lambda s: s.subject.identifier),
        ('duplicate_count_in_subject', lambda s:
            s.collapse.copy_number_in_subject))),


    'changeo': OrderedDict((
        ('SEQUENCE_ID', 'seq_id'),
        ('SEQUENCE_IMGT', 'sequence'),
        ('GERMLINE_IMGT_D_MASK', 'germline_d_masked'),
        ('FUNCTIONAL', 'functional'),
        ('IN_FRAME', 'in_frame'),
        ('STOP', 'stop'),
        ('V_CALL', lambda s: s.v_gene.replace(
           '|', ',' + GeneName(s.v_gene).prefix)),
        ('J_CALL', lambda s: s.j_gene.replace(
           '|', ',' + GeneName(s.j_gene).prefix)),
        ('JUNCTION_LENGTH', 'cdr3_num_nts'),
        ('JUNCTION', 'cdr3_nt'),
        ('V_IDENTITY', lambda s:
            round(100 * s.v_match / float(s.v_length), 2)),
        ('V_SCORE', 'v_match'),
        ('J_IDENTITY', lambda s:
            round(100 * s.j_match / float(s.j_length), 2)),
        ('J_SCORE', 'j_match'),
        ('DUPCOUNT', lambda s: s.copy_number),
        ('DUPCOUNT_IN_SUBJECT', lambda s: s.collapse.copy_number_in_subject),
        ('CLONE', 'clone_id'),

        # Extras
        ('CLONE_CDR3_NT', lambda s: s.clone.cdr3_nt if s.clone else ''),
        ('CLONE_CDR3_AA', lambda s: s.clone.cdr3_aa if s.clone else ''),
        ('SAMPLE', lambda s: s.sample.name),
        ('SUBJECT', lambda s: s.subject.identifier)))
}


class SequenceWriter(StreamingTSV):
    META_PREFIX = 'METADATA_'

    def __init__(self, format_name, metadata_fields):
        self.mapping = mappings[format_name]
        fields = list(self.mapping.keys()) + list(sorted([
            self.META_PREFIX + m for m in metadata_fields]))
        super(SequenceWriter, self).__init__(fields)

    def writeseq(self, seq):
        return super(SequenceWriter, self).writerow(self.format_seq(seq))

    def format_seq(self, seq):
        fields = {
            out_field: self._get_val(seq, model_field)
            for out_field, model_field in self.mapping.items()
        }

        fields.update({
            self.META_PREFIX + k: v
            for k, v in seq.sample.metadata_dict.items()
        })
        return fields

    def _get_val(self, model, field):
        if type(field) == str:
            val = getattr(model, field)
            if type(val) == bool:
                return 'T' if val else 'F'
            return val
        if callable(field):
            return field(model)
        return ''


def get_sequences(session, seqs, format_name):
    seqs = seqs.join(SequenceCollapse)
    seqs = seqs.options(
        joinedload(Sequence.clone),
        joinedload(Sequence.collapse),
        joinedload(Sequence.sample),
        joinedload(Sequence.subject),
    )
    meta_keys = set([m.key for m in session.query(SampleMetadata.key)])
    writer = SequenceWriter(format_name, meta_keys)

    yield writer.writeheader()
    for seq in yield_limit(seqs, Sequence.ai):
        yield writer.writeseq(seq)


def write_sequences(session, args):
    for subject in session.query(Subject):
        logger.info('Exporting subject {}'.format(subject.identifier))
        seqs = session.query(Sequence).filter(
            Sequence.subject_id == subject.id
        )
        if args.clones_only:
            seqs = seqs.filter(~Sequence.clone_id.is_(None))
        if args.min_subject_copies is not None:
            seqs = seqs.filter(
                SequenceCollapse.copy_number_in_subject >=
                args.min_subject_copies
            )

        fn = '{}.{}.tsv'.format(subject.identifier, args.fmt)
        write_tsv(fn, get_sequences, session, seqs, args.fmt)



import math
import numpy as np


def choose(n, k):
    return n ** k / math.factorial(k)


def hypergeom(length, mutation, K):
    M = length
    n = K
    N = np.ceil(length * mutation)

    def pmf(k):
        if M == K or k > N:
            return 0
        return (
            choose(n, k) * choose(M - n, N - k) /
            choose(M, N)
        )
    return np.sum(
        [pmf(k) * np.power(.33, k)
            for k in range(int(np.ceil(K / 2)), K)]
    )



from .sequences import get_sequences, write_sequences, mappings  # noqa: F401
from .genbank import write_genbank  # noqa: F401, F403
from .vdjtools import *  # noqa: F401, F403
from .selection import *  # noqa: F401, F403
from .samples import *  # noqa: F401, F403
from .clones import *  # noqa: F401, F403



import functools
import multiprocessing as mp
import traceback
import logging
import time

from immunedb.util.log import logger


class Worker(object):
    def log(self, lvl, msg):
        logger.log(lvl, 'Worker {}: {}'.format(self._worker_id, msg))

    def info(self, msg):
        self.log(logging.INFO, msg)

    def error(self, msg):
        self.log(logging.ERROR, msg)

    def warning(self, msg):
        self.log(logging.WARNING, msg)

    def do_task(self, args):
        raise NotImplementedError

    def cleanup(self):
        pass


class TaskQueue(object):
    def __init__(self):
        self._task_queue = mp.JoinableQueue()
        self._num_tasks = 0
        self._workers = []

    def add_task(self, args):
        self._num_tasks += 1
        self._task_queue.put(args)

    def add_tasks(self, tasks):
        for i, task in enumerate(tasks):
            self.add_task(task)

    def add_worker(self, worker):
        self._workers.append(
            mp.Process(
                target=self._func_wrap,
                args=(len(self._workers) + 1, worker)
            )
        )

    def add_workers(self, count, worker_cls, *args, **kwargs):
        for _ in range(count):
            self.add_worker(worker_cls(*args, **kwargs))

    def join(self):
        self._task_queue.join()

    def signal_end(self):
        for _ in self._workers:
            self.add_task(None)

    def start(self, block=True):
        if block:
            self.signal_end()

        for worker in self._workers:
            worker.start()
        if block:
            self.join()

    def _func_wrap(self, worker_id, worker):
        worker._worker_id = worker_id
        while True:
            try:
                args = self._task_queue.get()
                if args is None:
                    self._task_queue.task_done()
                    break
                else:
                    worker.do_task(args)
                    self._task_queue.task_done()
            except Exception:
                worker.error(
                    'The task was not completed because:\n{}'.format(
                        traceback.format_exc()))
                self._task_queue.task_done()
        worker.cleanup()

    def num_tasks(self):
        return self._num_tasks


def subcaller(func, data, i):
    return func(data[i])


# V2 of multiprocessing
def process_data(input_data, process_func, aggregate_func, nproc,
                 generate_args={}, process_args={}, aggregate_args={}):
    if callable(input_data):
        start = time.time()
        input_data = input_data(**generate_args)
        logger.info('Generate time: {}'.format(time.time() - start))

    with mp.Manager() as manager:
        proxy_data = manager.list(input_data)
        pool = mp.Pool(processes=nproc)
        f = functools.partial(
            subcaller,
            functools.partial(process_func, **process_args),
            proxy_data
        )
        start = time.time()
        logger.info('Waiting on pool {}'.format(process_func.__name__))

        res = [r for r in pool.map(f, range(len(proxy_data))) if r is not None]
        pool.close()
    logger.info('Pool done: {}'.format(time.time() - start))

    start = time.time()
    logger.info('Waiting on aggregation {}'.format(aggregate_func.__name__))
    ret = aggregate_func(res, **aggregate_args)
    logger.info('Done aggregation: {}'.format(time.time() - start))

    return ret



from sqlalchemy import func

from immunedb.common.models import CloneStats, Sample, SampleMetadata
from immunedb.exporting.tsv_writer import StreamingTSV, write_tsv
from immunedb.util.log import logger


class Passthrough:
    def __getattr__(self, attr):
        return 0


def get_samples(session, for_update=False):
    meta = [
        s.key for s in session.query(SampleMetadata.key).group_by(
            SampleMetadata.key).order_by(SampleMetadata.key)
    ]

    clone_cnts = {s.sample_id: s.clones for s in session.query(
        CloneStats.sample_id,
        func.count(CloneStats.clone_id.distinct()).label('clones')
    ).filter(
        ~CloneStats.sample_id.is_(None)
    ).group_by(CloneStats.sample_id)}

    if for_update:
        fields = ['name', 'new_name']
    else:
        fields = ['id', 'name', 'subject', 'input_sequences', 'identified',
                  'in_frame', 'stops', 'functional', 'clones']
    fields.extend(meta)
    writer = StreamingTSV(fields)
    yield writer.writeheader()
    for sample in session.query(Sample).order_by(Sample.name):
        row = {
            'id': sample.id,
            'name': sample.name,
            'new_name': sample.name
        }
        stats = sample.stats if sample.stats else Passthrough()
        if not for_update:
            row.update({
                'subject': sample.subject.identifier,
                'input_sequences': stats.sequence_cnt +
                stats.no_result_cnt,
                'identified': stats.sequence_cnt,
                'in_frame': stats.in_frame_cnt,
                'stops': stats.stop_cnt,
                'functional': stats.functional_cnt,
                'clones': clone_cnts.get(sample.id, 0)
            })

        row.update(sample.metadata_dict)
        yield writer.writerow(row)


def write_samples(session, args):
    logger.info('Exporting samples')
    write_tsv('samples.tsv', get_samples, session, args.for_update)



import json

from sqlalchemy import distinct, func

import immunedb.common.config as config
from immunedb.common.models import (Clone, CloneStats, Sequence,
                                    SequenceCollapse)
import immunedb.common.modification_log as mod_log
from immunedb.common.mutations import CloneMutations
import immunedb.util.concurrent as concurrent
from immunedb.util.log import logger


class CloneStatsWorker(concurrent.Worker):
    """A worker class for generating clone statistics.  This worker will accept
    one clone at a time for parallelization.

    :param Session session: The database session

    """
    def __init__(self, session):
        self._session = session

    def do_task(self, clone_id):
        """Starts the task of generating clone statistics for a given
        clone_id.
        :param int args: The clone_id for which to calculate statistics

        """

        existing = self._session.query(CloneStats).filter(
            CloneStats.clone_id == clone_id).first()

        if existing is not None:
            return

        self.info('Clone {}'.format(clone_id))
        sample_ids = [c.sample_id for c in self._session.query(
                distinct(Sequence.sample_id).label('sample_id')
            ).filter(
                Sequence.clone_id == clone_id
            )]
        sample_ids.append(None)
        for sample_id in sample_ids:
            self._process_sample(clone_id, sample_id)

    def _process_sample(self, clone_id, sample_id):
        """Processes clone statistics for one sample (or the aggregate of all
        samples).  If ``sample_id`` is None the statistics for all sequences in
        the clone is generated.

        :param int clone_id: The ID of the clone
        :param int sample_id: The ID of a sample in which the clone exists

        """

        if sample_id is None:
            counts = self._session.query(
                func.count(Sequence.ai).label('unique'),
                func.sum(SequenceCollapse.copy_number_in_subject).label(
                    'total')
            ).join(SequenceCollapse).filter(
                Sequence.clone_id == clone_id,
                SequenceCollapse.copy_number_in_subject > 0
            ).first()
        else:
            counts = self._session.query(
                func.count(Sequence.ai).label('unique'),
                func.sum(Sequence.copy_number).label('total')
            ).filter(
                Sequence.sample_id == sample_id,
                Sequence.clone_id == clone_id
            ).first()

        clone_inst = self._session.query(Clone).filter(
            Clone.id == clone_id).first()
        sample_mutations = CloneMutations(
            self._session,
            clone_inst
        ).calculate(
            commit_seqs=sample_id is not None, limit_samples=[sample_id],
        )[sample_id]

        record_values = {
            'clone_id': clone_id,
            'subject_id': clone_inst.subject_id,
            'functional': clone_inst.functional,
            'unique_cnt': counts.unique,
            'total_cnt': counts.total,
            'mutations': json.dumps(sample_mutations.get_all()),
        }

        self._session.add(CloneStats(sample_id=sample_id, **record_values))

        # If this clone only appears in one sample, the 'total clone' stats are
        # the same as for the single sample
        if sample_id is None:
            clone_inst.overall_unique_cnt = counts.unique
            clone_inst.overall_instance_cnt = self._session.query(
                func.count(Sequence.ai).label('instances'),
            ).filter(
                Sequence.clone_id == clone_id
            ).first().instances

            clone_inst.overall_total_cnt = counts.total
        self._session.commit()

    def cleanup(self):
        self._session.commit()
        self._session.close()


def run_clone_stats(session, args):
    """Runs the clone statistics generation stage of the pipeline.
    :param Session session: The database session
    :param Namespace args: The arguments passed to the command

    """
    mod_log.make_mod('clone_stats', session=session, commit=True,
                     info=vars(args))

    if args.clone_ids is not None:
        clones = args.clone_ids
    elif args.subject_ids is not None:
        clones = [c.id for c in session.query(Clone.id).filter(
            Clone.subject_id.in_(args.subject_ids))]
    else:
        clones = [c.id for c in session.query(Clone.id)]
    clones.sort()

    if args.regen:
        logger.info('Deleting old clone statistics for {} clones'.format(
            len(clones)))
        session.query(CloneStats).filter(
            CloneStats.clone_id.in_(clones)
        ).delete(synchronize_session=False)
        session.commit()

    tasks = concurrent.TaskQueue()
    logger.info('Creating task queue to generate stats for {} clones.'.format(
        len(clones)))
    for cid in clones:
        tasks.add_task(cid)

    for i in range(0, args.nproc):
        session = config.init_db(args.db_config)
        tasks.add_worker(CloneStatsWorker(session))

    tasks.start()



import json
from functools import partial, wraps
import os
import re
import signal
import time

from sqlalchemy.orm import scoped_session

import bottle
from bottle import response, request

try:
    from rollbar.contrib.bottle import RollbarBottleReporter
    ROLLBAR_SUPPORT = True
except ImportError:
    ROLLBAR_SUPPORT = False

import immunedb.api.queries as queries
from immunedb.common.models import Sequence, SequenceCollapse
from immunedb.util.log import logger
import immunedb.exporting as exporting


class EnableCors(object):
    """A class to enable Cross-Origin Resource Sharing to facilitate AJAX
    requests.

    """
    name = 'enable_cors'
    api = 2

    def apply(self, fn, context):
        def _enable_cors(*args, **kwargs):
            response.headers['Access-Control-Allow-Origin'] = '*'
            response.headers['Access-Control-Allow-Methods'] = ('GET, POST')
            response.headers['Access-Control-Allow-Headers'] = (
                'Origin, '
                'Accept, Content-Type, '
                'X-Requested-With, X-CSRF-Token')

            if bottle.request.method != 'OPTIONS':
                return fn(*args, **kwargs)

        return _enable_cors


def create_response(j=None, code=200, ctype='application/json'):
    if j is None:
        bottle.response.status = code
        return
    bottle.response.content_type = ctype
    bottle.response.status = code
    if ctype == 'application/json':
        return json.dumps(j)
    return j


def get_paging():
    data = bottle.request.json or {}
    if data.get('nopage', False):
        return None
    return [
        int(data.get('page', 1)),
        int(data.get('per_page', 10))
    ]


def decode_run_length(encoding):
    ids = []
    offset = 1
    for match in re.finditer('(T|F)(\d+)', encoding.upper()):
        size = int(match.group(2))
        if match.group(1) == 'T':
            ids.extend(range(offset, offset + size))
        offset += size
    return ids


def create_app(session_maker, allow_shutdown=False):
    app = bottle.Bottle()
    app.install(EnableCors())

    def with_session(f):
        @wraps(f)
        def _wrapper(*args, **kwargs):
            session = scoped_session(session_maker)()
            try:
                return f(session, *args, **kwargs)
            except Exception:
                raise
            finally:
                session.close()

        return _wrapper

    @app.route('/samples/list', method=['POST', 'OPTIONS'])
    @with_session
    def samples_list(session):
        return create_response(queries.get_samples(session))

    @app.route('/sequences/list', method=['POST', 'OPTIONS'])
    @with_session
    def sequences_list(session):
        fields = bottle.request.json or {}
        return create_response(queries.get_sequences(
            session,
            fields.get('filters', {}),
            fields.get('order_field', None),
            fields.get('order_dir', 'asc'),
            fields.get('subject_id', None),
            get_paging()))

    @app.route('/sequence/<sample_id>/<seq_id>', method=['POST', 'OPTIONS'])
    @with_session
    def sequence(session, sample_id, seq_id):
        return create_response(queries.get_sequence(session, sample_id,
                                                    seq_id))

    @app.route('/clones/list', method=['POST', 'OPTIONS'])
    @app.route('/clones/list/<subject_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clones_list(session, subject_id=None):
        fields = bottle.request.json or {}
        return create_response(queries.get_clones(
            session,
            fields.get('filters', {}),
            fields.get('order_field', None),
            fields.get('order_dir', 'asc'),
            subject_id,
            get_paging()))

    @app.route('/clone/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone(session, clone_id):
        return create_response(queries.get_clone(session, clone_id))

    @app.route('/clone/sequences/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_sequences(session, clone_id):
        fields = bottle.request.json or {}
        return create_response(queries.get_clone_sequences(
            session, clone_id, fields.get('get_collapse', False), get_paging()
        ))

    @app.route('/clone/mutations/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_mutations(session, clone_id):
        fields = bottle.request.json or {}
        return create_response(queries.get_clone_mutations(
            session,
            clone_id,
            fields.get('type', 'percent'),
            int(fields.get('value', 0))
        ))

    @app.route('/clone/lineage/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_lineage(session, clone_id):
        return create_response(queries.get_clone_tree(session, clone_id))

    @app.route('/clone/pressure/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_pressure(session, clone_id):
        return create_response(list(
            queries.get_selection_pressure(session, clone_id)))

    @app.route('/samples/analyze/<sample_encoding>', method=[
        'POST', 'OPTIONS'])
    @with_session
    def analyze_samples(session, sample_encoding):
        fields = bottle.request.json or {}
        return create_response(
            queries.analyze_samples(
                session,
                decode_run_length(sample_encoding),
                fields.get('filter_type', 'unique_multiple'),
                fields.get('include_outliers', True),
                fields.get('include_partials', True),
                fields.get('percentages', False),
                fields.get('grouping', 'name'),
            )
        )

    @app.route('/samples/overlap/<sample_encoding>', method=[
        'POST', 'OPTIONS'])
    @with_session
    def overlap(session, sample_encoding):
        fields = bottle.request.json or {}
        return create_response(queries.get_clone_overlap(
            session,
            decode_run_length(sample_encoding),
            fields.get('filter_type', 'clones_all'),
            get_paging())
        )

    @app.route('/samples/v_usage/<sample_encoding>', method=[
        'POST', 'OPTIONS'])
    @with_session
    def v_usage(session, sample_encoding):
        fields = bottle.request.json or {}
        data, x_categories, totals, prefix = queries.get_v_usage(
            session,
            decode_run_length(sample_encoding),
            fields.get('filter_type', 'unique_multiple'),
            fields.get('include_outliers', True),
            fields.get('include_partials', True),
            fields.get('grouping', 'name'),
            fields.get('by_family', False)
        )

        x_categories.sort()
        y_categories = sorted(data.keys())
        array = []
        min_v = 100
        max_v = 0
        for i, x in enumerate(x_categories):
            for j, y in enumerate(y_categories):
                usage_for_y = data[y]
                if x in usage_for_y:
                    min_v = min(min_v, usage_for_y[x])
                    max_v = max(max_v, usage_for_y[x])
                    array.append([i, j, usage_for_y[x]])
                else:
                    array.append([i, j, 0])

        return create_response({
            'x_categories': ['{}{}'.format(prefix, e) for e in x_categories],
            'y_categories': y_categories,
            'totals': totals,
            'data': array,
            'min': min_v,
            'max': max_v
        })

    @app.route('/subjects/list', method=['POST', 'OPTIONS'])
    @with_session
    def subjects_list(session):
        return create_response(queries.get_all_subjects(session, get_paging()))

    @app.route('/subject/<subject_id>', method=['POST', 'OPTIONS'])
    @with_session
    def subject(session, subject_id):
        return create_response(queries.get_subject(session, subject_id))

    def set_file(suffix, ext='tsv'):
        time_str = time.strftime('%Y-%m-%d-%H-%M')
        fn = '{}_{}.{}'.format(time_str, suffix, ext)
        response.headers['Content-Disposition'] = 'attachment;filename=' + fn

    @app.route('/export/sequences/<schema>', method=['GET', 'OPTIONS'])
    @with_session
    def export_sequences(session, schema):
        if schema not in exporting.mappings:
            return create_response(code=400)

        seqs = session.query(Sequence)

        min_copies = request.query.get('min_subject_copies', 0)
        if min_copies:
            seqs = seqs.filter(
                SequenceCollapse.copy_number_in_subject >= min_copies
            )

        if request.query.get('clones_only', False):
            seqs = seqs.filter(~Sequence.clone_id.is_(None))

        set_file(schema)

        yield from exporting.get_sequences(session, seqs, schema)

    @app.route('/export/clones/<schema>', method=['GET', 'OPTIONS'])
    @with_session
    def export_clones(session, schema):
        funcs = {
            'summary': partial(
                exporting.get_clone_summary,
                include_lineages=request.query.get('lineages', False)
            ),
            'overlap': exporting.get_clone_overlap,
            'selection': exporting.get_selection,
            'vdjtools': exporting.get_vjdtools_as_zip,
        }
        if schema not in funcs:
            return create_response(code=400)

        set_file('clone_' + schema,
                 'zip' if schema == 'vdjtools' else 'tsv')

        yield from funcs[schema](session)

    @app.route('/export/samples', method=['GET', 'OPTIONS'])
    @with_session
    def export_samples(session):
        set_file('samples')
        yield from exporting.get_samples(session)

    @app.route('/shutdown', method=['GET'])
    def shutdown():
        if allow_shutdown:
            logger.warning('Shutting down from remote request')
            os.kill(os.getppid(), signal.SIGINT)
        return create_response(code=404)

    return app


def run_rest_service(session_maker, args):
    if args.rollbar_token:
        if not ROLLBAR_SUPPORT:
            logger.error('Rollbar is not installed')
            return
        rbr = RollbarBottleReporter(
            access_token=args.rollbar_token,
            environment=args.rollbar_env)
        bottle.install(rbr)

    app = create_app(session_maker, args.allow_shutdown)
    if args.debug:
        app.catchall = False
    app.run(host='0.0.0.0', port=args.port, server=args.server,
            debug=args.debug, worker_class='eventlet', timeout=0)



import json

from sqlalchemy import distinct

from immunedb.common.models import Sequence, SequenceCollapse
import immunedb.util.lookups as lookups
import immunedb.util.funcs as funcs


class ContextualMutations(object):
    """Calculates the mutations of a set of sequences within a given
    context.

    :param list regions: The gene region positions relative to the input
        sequence gapping.

    """

    def __init__(self, regions):
        self._seen = {}
        self._regions = regions
        self._pos_seen = set([])
        self.region_muts = {}
        self.position_muts = {}

    def add_mutation(self, seq, cdr3_num_nts, mutation, from_aa,
                     intermediate_seq_aa, final_seq_aa, copy_number):
        """Adds a mutation to the the aggregate mutation list.

        :param str seq: The sequence with the mutation
        :param int cdr3_num_nts: The number of bases in the CDR3
        :param tuple mutation: The mutation in (position, from_nt, to_nt, type)
            form.
        :param char from_aa: The germline amino acid
        :param char intermediate_seq_aa: The amino acid in the sequence if only
            the point mutation occurred
        :param char final_seq_aa: The final mutated amino acid
        :param int copy_number: The number of times this sequence appeared

        """
        pos, _, _, mtype = mutation
        region = funcs.get_pos_region(self._regions, cdr3_num_nts, pos)
        self._add_to_region(seq, cdr3_num_nts, mutation, from_aa,
                            intermediate_seq_aa, final_seq_aa, copy_number,
                            region)
        self._add_to_region(seq, cdr3_num_nts, mutation, from_aa,
                            intermediate_seq_aa, final_seq_aa, copy_number,
                            'ALL')

        if pos not in self.position_muts:
            self.position_muts[pos] = {}
        if mtype not in self.position_muts[pos]:
            self.position_muts[pos][mtype] = 0
        self.position_muts[pos][mtype] += 1

    def _add_to_region(self, seq, cdr3_num_nts, mutation, from_aa,
                       intermediate_seq_aa, final_seq_aa, copy_number, region):
        pos, from_nt, to_nt, mtype = mutation
        if region not in self.region_muts:
            self.region_muts[region] = {}

        # If it's a new mutation, setup the dictionaries
        if mtype not in self.region_muts[region]:
            self.region_muts[region][mtype] = {}

        if mutation not in self.region_muts[region][mtype]:
            self.region_muts[region][mtype][mutation] = {
                'pos': pos,
                'from_nt': from_nt,
                'from_aa': from_aa,
                'to_nt': to_nt,
                'to_aas': [],

                'unique': 0,
                'total': 0,
                'intermediate_aa': intermediate_seq_aa,
            }

        mut_dict = self.region_muts[region][mtype][mutation]
        if final_seq_aa not in mut_dict['to_aas']:
            mut_dict['to_aas'].append(final_seq_aa)

        mut_dict['unique'] += 1
        mut_dict['total'] += copy_number

    def get_all(self):
        # Strip the dictionary keys and just make a list of mutations
        final_regions = {}
        for region, types in self.region_muts.items():
            final_regions[region] = {}
            for mtype, mutations in types.items():
                final_regions[region][mtype] = list(mutations.values())

        return {
            'regions': final_regions,
            'positions': self.position_muts
        }


class CloneMutations(object):
    def __init__(self, session, clone):
        self._clone = clone
        self._session = session
        self._germline = self._clone.consensus_germline

    def _get_codon_at(self, seq, i):
        aa_off = i - i % 3
        return seq[aa_off:aa_off + 3]

    def _get_aa_at(self, seq, i):
        return lookups.aa_from_codon(self._get_codon_at(seq, i))

    def _get_mutation(self, seq, i):
        if (self._germline[i] != seq[i] and
                self._germline[i] not in ('N', '-') and
                seq[i] not in ('N', '-')):
            grm_aa = self._get_aa_at(self._germline, i)
            # Simulate this mutation alone
            off = i % 3
            grm_codon = self._get_codon_at(self._germline, i)
            seq_aa = lookups.aa_from_codon(
                grm_codon[:off] + seq[i] + grm_codon[off+1:])

            if grm_aa is None or seq_aa is None:
                return 'unknown', seq_aa
            elif grm_aa != seq_aa:
                if lookups.are_conserved_aas(grm_aa, seq_aa):
                    return 'conservative', seq_aa
                return 'nonconservative', seq_aa
            else:
                return 'synonymous', seq_aa

        return None, None

    def calculate(self, commit_seqs=False, limit_samples=None):
        sample_mutations = {}

        if limit_samples is not None:
            sample_ids = limit_samples
        else:
            sample_ids = [r.sample_id for r in self._session.query(
                distinct(Sequence.sample_id).label('sample_id')
                ).filter(
                Sequence.clone == self._clone
                )]
            sample_ids.append(None)

        for sample_id in sample_ids:
            seqs = self._session.query(Sequence).filter(
                Sequence.clone == self._clone)
            if sample_id is None:
                seqs = seqs.join(SequenceCollapse).filter(
                    SequenceCollapse.copy_number_in_subject > 0
                )
            else:
                seqs = seqs.filter(
                    Sequence.sample_id == sample_id
                )
            sample_mutations[sample_id] = self._get_contextual_mutations(
                seqs, commit_seqs, use_sample_copy=sample_id is not None)

        return sample_mutations

    def _get_contextual_mutations(self, seqs, commit_seqs, use_sample_copy):
        context_mutations = ContextualMutations(self._clone.regions)
        for seq in seqs:
            seq_mutations = {}
            for i in range(0, len(seq.clone_sequence)):
                mtype, intermediate_aa = self._get_mutation(
                    seq.clone_sequence, i)
                if mtype is None:
                    continue

                from_aa = self._get_aa_at(self._germline, i)
                seq_mutations[i] = mtype

                mutation = (i, self._germline[i], seq.clone_sequence[i], mtype)
                copy_field = (
                    seq.copy_number if use_sample_copy else
                    seq.collapse.copy_number_in_subject
                )
                context_mutations.add_mutation(
                    seq.clone_sequence, self._clone.cdr3_num_nts,
                    mutation, from_aa, intermediate_aa,
                    self._get_aa_at(seq.clone_sequence, i), copy_field)

            if commit_seqs:
                seq.mutations_from_clone = json.dumps(seq_mutations)
        return context_mutations


def threshold_mutations(all_muts, min_required_seqs):
    """Removes mutations that occur in less than ``min_required_seqs``

    :param dict all_muts: A mutation dictionary as generated by
        ContextualMutations
    :param int min_required_seqs: The minimum number of sequences in which a
        mutation must occur to be in the thresholded mutations

    :returns dict: A new mutation dictionary with infrequent mutations removed

    """
    final = {}
    for region, types in all_muts['regions'].items():
        final[region] = {
            'counts': {
                'total': {},
                'unique': {}
            },
            'mutations': {}
        }
        for mtype, mutations in sorted(types.items()):
            for mutation in sorted(
                    mutations,
                    key=lambda m: (m['pos'], m['from_nt'], m['to_nt'])):
                if mutation['unique'] >= min_required_seqs:
                    if mtype not in final[region]['mutations']:
                        final[region]['mutations'][mtype] = []
                    if mtype not in final[region]['counts']['total']:
                        final[region]['counts']['total'][mtype] = 0
                        final[region]['counts']['unique'][mtype] = 0
                    final[region]['mutations'][mtype].append(mutation)
                    final[region]['counts']['total'][mtype] += \
                        mutation['unique']
                    final[region]['counts']['unique'][mtype] += 1
    return final



import datetime
import json

from immunedb.common.models import ModificationLog


class LoggingException(Exception):
    pass


def make_mod(action_type, info, session=None, commit=False):
    ml = ModificationLog(
        datetime=datetime.datetime.utcnow(),
        action_type=action_type,
        info=json.dumps(info))
    if session is None:
        if commit:
            raise LoggingException('Specified `commit` without `session`')
        return ml

    session.add(ml)
    if commit:
        session.commit()
    return ml



from sqlalchemy.orm import joinedload

from immunedb.common.models import SelectionPressure
from immunedb.exporting.tsv_writer import StreamingTSV, write_tsv
from immunedb.util.funcs import yield_limit
from immunedb.util.log import logger


def get_selection(session, filter_type=None):
    query = session.query(SelectionPressure).options(
        joinedload(SelectionPressure.clone),
        joinedload(SelectionPressure.sample),
    )
    if filter_type == 'overall':
        query = query.filter(SelectionPressure.sample_id.is_(None))
    elif filter_type == 'samples':
        query = query.filter(~SelectionPressure.sample_id.is_(None))

    base_fields = SelectionPressure.__table__.c.keys()
    base_fields.remove('id')
    base_fields.remove('sample_id')

    writer = StreamingTSV(['sample', 'subject'] + base_fields)
    yield writer.writeheader()

    for sel in yield_limit(query, SelectionPressure.id):
        row = {f: getattr(sel, f) for f in base_fields}
        row['sample'] = sel.sample.name if sel.sample else 'All Samples'
        row['subject'] = sel.clone.subject.identifier
        yield writer.writerow(row)


def write_selection(session, args):
    logger.info('Exporting selection pressure')
    write_tsv('selection_pressure.tsv', get_selection, session, args.filter)



import coloredlogs
import logging

logger = logging.getLogger('immunedb')
colors = coloredlogs.DEFAULT_FIELD_STYLES
colors['levelname']['color'] = 'white'
coloredlogs.install(level='DEBUG', logger=logger,
                    fmt='%(asctime)s [%(levelname)s] %(message)s',
                    field_styles=colors)
handler = logging.StreamHandler()



import csv
import os
import re

from sqlalchemy.sql import exists

from immunedb.common.models import Sample, Sequence
from immunedb.util.log import logger

REQUIRED_FIELDS = ('file_name', 'study_name', 'sample_name', 'subject')
COMMON_FIELDS = ('date', 'subset', 'tissue', 'disease', 'lab', 'experimenter',
                 'ig_class', 'timepoint', 'v_primer', 'j_primer')
NA_VALUES = ('na', 'n/a', 'null', 'none', '')


class MetadataException(Exception):
    pass


def check_populated(row):
    missing = [f for f in REQUIRED_FIELDS if f not in row]
    if len(missing) > 0:
        if 'sample_name' in missing:
            raise MetadataException('Sample name cannot be blank')
        raise MetadataException(
            'Fields {} cannot be blank for sample {}'.format(
                ','.join(missing), row['sample_name']))


def parse_metadata(session, fh, warn_existing, warn_missing, path):
    reader = csv.DictReader(fh, delimiter='\t')
    provided_fields = set(reader.fieldnames)
    for field in provided_fields:
        if not re.match('^[a-z][a-z0-9_]*$', field):
            raise MetadataException(
                'Metadata headers must only contain numbers, letters, and '
                'underscores, and cannot start with a number: {}'.format(field)
            )

    missing_fields = set(REQUIRED_FIELDS) - provided_fields
    if len(missing_fields) > 0:
        raise MetadataException(
            'Metadata is missing the following headers: {}'.format(
                ','.join(missing_fields)))

    metadata = {}
    for row in reader:
        row = {k: v for k, v in row.items()
               if v is not None and len(v) > 0 and v.lower() not in NA_VALUES}
        if len(row) == 0:
            continue
        check_populated(row)
        # Check if the sample name is unique
        if row['sample_name'] in metadata:
            raise MetadataException(
                'Duplicate sample name {} in metadata.'.format(
                    row['sample_name']))

        # Check if a sample with the same name is in the database
        sample_in_db = session.query(Sample).filter(
            Sample.name == row['sample_name'],
            exists().where(
                Sequence.sample_id == Sample.id
            )).first()
        if sample_in_db:
            message = 'Sample {} already exists. {}'.format(
                row['sample_name'],
                'Skipping.' if warn_existing else 'Cannot continue.'
            )
            if warn_existing:
                logger.warning(message)
                continue
            else:
                raise MetadataException(message)

        # Check if specified file exists
        if not os.path.isfile(os.path.join(path, row['file_name'])):
            message = (
                'File {} for sample {} does not exist. {}'.format(
                    row['file_name'], row['sample_name'],
                    'Skipping.' if warn_missing else 'Cannot continue.'))
            if warn_missing:
                logger.warning(message)
                continue
            else:
                raise MetadataException(message)

        metadata[row['sample_name']] = row

    return metadata



import csv
import io


class StreamingTSV(object):
    def __init__(self, fieldnames):
        self.out = io.StringIO()
        self.fieldnames = fieldnames
        self.writer = csv.DictWriter(self.out, fieldnames=self.fieldnames,
                                     delimiter='\t', extrasaction='ignore')

    def writeheader(self):
        return self._stream(self.writer.writeheader)

    def writerow(self, row):
        return self._stream(self.writer.writerow, row)

    def _stream(self, action, *args, **kwargs):
        self.out.truncate(0)
        self.out.seek(0)
        action(*args, **kwargs)
        return self.out.getvalue().strip() + '\n'


def write_tsv(fn, func, *args, **kwargs):
    with open(fn, 'w+') as fh:
        for line in func(*args, **kwargs):
            fh.write(line)



import csv

from sqlalchemy.orm import joinedload

from immunedb.common.models import Clone, CloneStats, SampleStats, Sequence
from immunedb.aggregation.clones import generate_consensus, push_clone_ids
from immunedb.importing import ImportException


def generate_template(session, out_file):
    with open(out_file, 'w+') as fh:
        writer = csv.DictWriter(fh, delimiter='\t', fieldnames=[
            'ai',
            'seq_id',
            'sample_id',
            'subject_id',
            'subject',
            'sequence',
            'v_gene',
            'j_gene',
            'cdr3_aa',
            'cdr3_nt',
            'cdr3_num_nts',
            'copy_number',
            'copy_number_in_subject',
            'clone_id'
        ])

        writer.writeheader()
        template_seqs = session.query(Sequence).options(
            joinedload(Sequence.collapse)).order_by(Sequence.ai)
        for seq in template_seqs:
            if seq.collapse.copy_number_in_subject == 0:
                continue
            writer.writerow({
                'ai': seq.ai,
                'seq_id': seq.seq_id,
                'sample_id': seq.sample_id,
                'subject_id': seq.subject_id,
                'subject': seq.subject.identifier,
                'sequence': seq.sequence,
                'v_gene': seq.v_gene,
                'j_gene': seq.j_gene,
                'cdr3_aa': seq.cdr3_aa,
                'cdr3_nt': seq.cdr3_nt,
                'cdr3_num_nts': seq.cdr3_num_nts,
                'copy_number': seq.copy_number,
                'copy_number_in_subject': seq.collapse.copy_number_in_subject,
                'clone_id': 0
            })


def import_template(session, in_file, regen):
    if regen:
        session.query(Clone).delete()
        session.query(CloneStats).delete()
        session.query(SampleStats).delete()
        session.commit()

    seen_clones = {}
    with open(in_file) as fh:
        reader = csv.DictReader(fh, delimiter='\t')
        if ('ai' not in reader.fieldnames or
                'clone_id' not in reader.fieldnames):
            raise ImportException(
                'Input file must have "ai" and "clone_id" fields')
        for line in reader:
            if line['clone_id'] in ('None', '0', ''):
                continue
            clone_info = seen_clones.setdefault(line['clone_id'], {
                'clone': Clone(subject_id=int(line['subject_id']),
                               v_gene=line['v_gene'],
                               j_gene=line['j_gene'],
                               cdr3_num_nts=int(line['cdr3_num_nts'])),
                'seqs': []
            })
            clone_inst = clone_info['clone']
            if (clone_inst.v_gene != line['v_gene'] or
                    clone_inst.j_gene != line['j_gene'] or
                    clone_inst.cdr3_num_nts != int(line['cdr3_num_nts']) or
                    clone_inst.subject_id != int(line['subject_id'])):
                raise ImportException(
                    'Sequence with ai {} was assigned to clone {} with '
                    'mismatching v_gene, j_gene, cdr3_num_nts, or '
                    'subject.'.format(line['ai'], line['clone_id']))
            clone_info['seqs'].append({
                'ai': int(line['ai']),
                'sample_id': int(line['sample_id'])
            })
        session.commit()

    db_clone_ids = set([])
    for cid, clone_info in sorted(seen_clones.items()):
        clone_inst = clone_info['clone']
        session.add(clone_inst)
        session.flush()
        db_clone_ids.add(clone_inst.id)

        to_update = [{
            'sample_id': s['sample_id'],
            'ai': s['ai'],
            'clone_id': clone_inst.id
        } for s in clone_info['seqs']]
        session.bulk_update_mappings(Sequence, to_update)
    session.commit()
    generate_consensus(session, db_clone_ids)
    push_clone_ids(session)



from immunedb.common.models import Clone, CloneStats
from immunedb.exporting.tsv_writer import StreamingTSV, write_tsv
from immunedb.util.funcs import yield_limit
from immunedb.util.log import logger


def get_clone_summary(session, include_lineages):
    fields = [
        'clone_id', 'subject', 'v_gene', 'j_gene', 'functional', 'insertions',
        'deletions', 'cdr3_nt', 'cdr3_num_nt', 'cdr3_aa',
        'uniques', 'instances', 'copies', 'germline', 'parent_id',
        'avg_mutations_per_copy'
    ]
    if include_lineages:
        fields.append('lineage')
    writer = StreamingTSV(fields)

    yield writer.writeheader()
    for clone in yield_limit(session.query(Clone), Clone.id):
        row = {}
        for field in writer.fieldnames:
            try:
                row[field] = getattr(clone, field)
            except AttributeError:
                pass
        row.update({
            'clone_id': clone.id,
            'subject': clone.subject.identifier,
            'functional': 'T' if clone.functional else 'F',
            'insertions': clone._insertions,
            'deletions': clone._deletions,
            'uniques': clone.overall_unique_cnt,
            'instances': clone.overall_instance_cnt,
            'copies': clone.overall_total_cnt,
            'avg_mutations_per_copy': round(
                clone.overall_stats.total_mutations(normalize=True), 2)
        })
        if include_lineages:
            row['lineage'] = clone.tree
        yield writer.writerow(row)


def get_clone_overlap(session):
    writer = StreamingTSV(['clone_id', 'sample', 'uniques', 'copies',
                           'avg_mutations_per_copy'])

    stats = session.query(
        CloneStats
    ).filter(
        ~CloneStats.sample_id.is_(None)
    )

    yield writer.writeheader()
    for stat in yield_limit(stats, CloneStats.id):
        yield writer.writerow({
            'clone_id': stat.clone_id,
            'sample': stat.sample.name,
            'uniques': stat.unique_cnt,
            'copies': stat.total_cnt,
            'avg_mutations_per_copy': round(
                stat.total_mutations(normalize=True), 2)
        })


def write_clone_summary(session, args):
    logger.info('Exporting clone summary {} lineages'.format(
        'INCLUDING' if args.include_lineages else 'EXCLUDING'))
    write_tsv('clones.tsv', get_clone_summary, session, args.include_lineages)


def write_clone_overlap(session, args):
    logger.info('Exporting clone overlap')
    write_tsv('clone_overlap.tsv', get_clone_overlap, session)
